/*
 * WorldCanvas.java
 *
 * Created on September 14, 2005, 11:14 PM
 *
 * BallWorld for ENGR2510, fall 2005 at Olin, part of the CS101 project
 * (c)2005 Olin College
 */

package ballworld.ui;

import cs101.anim.Animate;
import cs101.anim.Animator;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Arc2D;
import java.util.HashSet;
import java.util.Set;
import javax.swing.JPanel;
import ballworld.*;

/**
 * A canvas for drawing Balls.  This class serves as the Balls' Worled, and
 * animates the Balls that implement the Animate interface.  It also gets user
 * input and forwards the events as appropriate to the selected Ball.
 *
 * @author Mike Wessler
 */
public class WorldCanvas extends JPanel
        implements World, Animate, MouseListener, MouseMotionListener,
        KeyListener
{
    /** the list of Balls */
    private Set<Ball> balls;
    
    /** the currently selected Ball, or null if no Ball is selected */
    private Ball selection;
    
    /** the background color */
    private Paint background = new Color(40,200,120); //new Color(20, 100, 60);
    
    /** the color of the selected Ball */
    private Paint selectedPaint = Color.red;
    
    /** the color of an unselected Ball */
    private Paint ballPaint = Color.black;
    
    /** the horizontal distance from the center of the selected ball to the
     * initial click location that selected it */
    private double xoffset;
    
    /** the vertical distance from the center of the selected ball to the
     * initial click location that selected it */
    private double yoffset;
    
    /** the Animator that animates the balls */
    private Animator anim;
    
    /** whethe the action is paused or not */
    private boolean paused = false;
    
    /** Creates a new instance of WorldCanvas */
    public WorldCanvas() {
        setPreferredSize(new Dimension(500, 500));
        balls = new HashSet<Ball>();
        setFocusable(true);
        addMouseListener(this);
        addMouseMotionListener(this);
        addKeyListener(this);
        anim = new Animator(this, Animator.StartOption.START_LATER, 0.05f);
    }

    /**
     * Removes all balls from the screen 
     */
    public void clear() {
        synchronized(balls) {
            balls.clear();
        }
    }
    
    /**
     * Sets whether the Animator thread is stopped or started.
     * @param paused true if the Animator thread should be stopped, false if
     * it should be restarted.
     */
    public void setPaused(boolean paused) {
        this.paused = paused;
        if (paused) {
            anim.stop();
        } else if (balls.size()>0) {
            anim.start();
        }
    }

    // javadoc from World
    public void removeBall(Ball ball) {
        if (selection == ball) {
            selection = null;
        }
        synchronized(balls) {
            balls.remove(ball);
        }
        if (paused) {
            repaint();
        } else if (balls.size()==0) {
            anim.stop();
            repaint();
        }
    }

    // javadoc from World
    public void addBall(Ball ball) {
        synchronized(balls) {
            balls.add(ball);
            ball.setWorld(this);
        }
        if (paused) {
            repaint();
        } else if (!anim.isAnimating()) {
            anim.start();
        }
    }

    // javadoc from World
    public double getMinY() {
        return -(double)getHeight()/2.0;
    }

    // javadoc from World
    public double getMinX() {
        return -(double)getWidth()/2.0;
    }

    // javadoc from World
    public double getMaxY() {
        return (double)getHeight()/2.0;
    }

    // javadoc from World
    public double getMaxX() {
        return (double)getWidth()/2.0;
    }

    // javadoc from World
    public Ball getClosestBall(Ball target) {
        int width = getWidth();
        int height = getHeight();
        // keep track of square of min distance to other ball
        double dist2= width*width+height*height;  // start big
        Ball best = null;
        synchronized(balls) {
            for (Ball ball : balls) {
                if (ball != target) {
                    double dx = ball.getX()-target.getX();
                    double dy = ball.getY()-target.getY();
                    double d2 = dx*dx+dy*dy;
                    if ((ball != target) && (d2 < dist2)) {
                        dist2 = d2;
                        best = ball;
                    }
                }
            }
        }
        return best;
    }

    // javadoc from Animate
    public void act() {
        Object ballArr[];
        synchronized(balls) {
            ballArr = balls.toArray();
        }
        for (Object elt : ballArr) {
            if (elt instanceof Animate) {
                Animate animate = (Animate)elt;
                animate.act();
            }
        }
        repaint();
    }
    
    /**
     * paints the board
     */
    public void paint(Graphics g1) {
        Graphics2D g = (Graphics2D)g1;
        requestFocusInWindow();
        
        g.setPaint(background);
        g.fillRect(0, 0, getWidth(), getHeight());
        Ball[] list;
        synchronized(balls) {
            list = new Ball[balls.size()];
            list = (Ball[])balls.toArray(list);
        }
        for (Ball ball : list) {
            double x = ball.getX()+getWidth()/2.0;
            double y = ball.getY()+getHeight()/2.0;
            double r = ball.getRadius();
            Arc2D arc = new Arc2D.Double(x-r, y-r, r*2, r*2, 0, 360,
                    Arc2D.CHORD);
            if (ball == selection) {
                g.setPaint(selectedPaint);
            } else {
                g.setPaint(ballPaint);
            }
            g.fill(arc);
        }
    }

    // javadoc from KeyListener
    public void keyTyped(KeyEvent evt) {
        if (selection != null) {
            selection.userTyped(evt.getKeyChar());
        }
    }

    // javadoc from KeyListener
    public void keyReleased(KeyEvent evt) {
    }

    // javadoc from KeyListener
    public void keyPressed(KeyEvent evt) {
    }

    // javadoc from MouseListener
    public void mouseReleased(MouseEvent evt) {
        mouseDragged(evt);
    }

    // javadoc from MouseListener
    public void mousePressed(MouseEvent evt) {
        // find the ball we hit
        Ball found = null;
        synchronized(balls) {
            for (Ball ball : balls) {
                double dx = evt.getX()-getWidth()/2.0 - ball.getX();
                double dy = evt.getY()-getHeight()/2.0 - ball.getY();
                double d2 = dx*dx + dy*dy;
                double checkr = ball.getRadius()+2;
                if (d2 < checkr*checkr) {
                    found = ball;
                    xoffset = -dx;
                    yoffset = -dy;
                }
            }
        }
        if (found != selection) {
            selection = found;
            if (paused) {
                repaint();
            }
        }
    }

    // javadoc from MouseListener
    public void mouseClicked(MouseEvent evt) {
    }
    
    // javadoc from MouseListener
    public void mouseExited(MouseEvent evt) {
    }

    // javadoc from MouseListener
    public void mouseEntered(MouseEvent evt) {
    }

    // javadoc from MouseMotionListener
    public void mouseDragged(MouseEvent evt) {
        if (selection != null) {
            double x = evt.getX()-(double)getWidth()/2.0 + xoffset;
            double y = evt.getY()-(double)getHeight()/2.0 + yoffset;
            selection.userDragged(x, y);
            repaint();
        }
    }

    // javadoc from MouseMotionListener
    public void mouseMoved(MouseEvent evt) {
    }

}
